#lang typed/racket

(provide (all-defined-out))

(define-type (stream T)
  (-> (stream-add T))
)

(struct [T] stream-add (
  [first : T]
  [rest : (stream T)]
  )
  #:transparent
)







(: powers-of-two (stream Number))
(define (powers-of-two)
  #|
    def powers_of_two():
      count = 1
      while True:
        yield count
        count = count * 2
  |#
  (: loop (-> Number (stream Number)))
  (define (loop count)
    (displayln ".")
    (thunk
      (stream-add count (loop (* count 2)))
    )
  )
  ((loop 1))
)





(: sample3
  (All [T]
    (->
      (stream T)
      (Listof T)
    )
  )
)
(define (sample3 s)
  (match (s)
    [(stream-add x s)
      (match (s)
        [(stream-add y s)
          (match (s)
            [(stream-add z _)
              (list x y z)
            ]
          )
        ]
      )
    ]
  )
)
;; (sample3 powers-of-two)



(: count-until
  (All [T]
    (->
      ; Given a predicate
      (-> T Boolean)
      ; Take a stream
      (stream T)
      ; How many elements
      Number
    )
  )
)

#|

count = 0
for x in stream:
  if predicate(x):
    count += 1
    # loop
  else:
    return count
|#

(define (count-until predicate s)
  (: loop
    (-> Number (stream T) Number)
  )
  (define (loop count s)
    (match (s)
      [(stream-add x s)
        (if (predicate x)
          (loop (+ count 1) s)
          count
        )
      ]
    )
  )
  (loop 0 s)
)

;; (count-until
;;   (lambda ([ x : Number])
;;     #t
;;   )
;;   powers-of-two
;; )

(: const (-> Number (stream Number)))
(define (const k)
  (thunk
    (stream-add k (const k) )
  )
)

;(sample3 (const 3))

(: naturals (stream Number))
(define (naturals)
  (: loop (-> Number (stream Number)))
  (define (loop num)
    (thunk
      (stream-add num
        (loop (+ num 1))
      )
    )
  )
  ((loop 0))
)

(: stream-map
  (All [InputElem OutputElem]
    (->
      ; Can convert an InputElem into an OutputElem
      (-> InputElem OutputElem)
      ; Given a stream of InputElem
      (stream InputElem)
      ; Generate a stream of OutputElem
      (stream OutputElem)
    )
  )
)
(define (stream-map f s)
  (thunk
    (match (s)
      [(stream-add x s)
        (stream-add
          (f x)
          (stream-map f s)
        )
      ]
    )
  )
)
#|
(sample3
(stream-map
  (lambda ([x : Number])
    (displayln x)
    (* x 2))
  naturals
)
)|#

(: trap (stream Number))
(define (trap)
  (error "You tripped the trap!")
)

(sample3
(stream-map
  (lambda ([x : Number])
    (displayln x)
    (* x 2))
  trap
)
)
